--------------------------------------------------------------------------
-- File vhdl/eu_inc/eu_cmp.vhdl - CMP Unit Execution
-- Copyright (C) 2002 Etienne Labarre (etienne.labarre@gadz.org)
--
-- Sun Jul 28 13:34:02 CEST 2002 : E. LABARRE
--  First version
--
--------------------------BEGIN-VHDL-LICENCE-----------------------------
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
---------------------------END-VHDL-LICENCE------------------------------
--
-- SIMD MODES
-- see file find_lsb.vhdl
-- 
-- OPERATIONS MODE
-- eu_inc_mode operation test
-- 000      CMP       
-- 001      MIN       
-- 010      MAX       
-- 011      SORT
-- 100      MSB0
-- 101      MSB1
--
--------------------------------------------------------------------------

-- include standard libraries
LIBRARY ieee;
USE ieee.std_logic_1164.ALL;
USE ieee.numeric_std.all;

-- include project libraries
LIBRARY work;
USE work.find_lsb;

-- Definition of EU_CMP entity
entity eu_cmp is
   generic ( eu_cmp_width : natural := 6 );
   port (
      eu_cmp_A    : in  std_ulogic_vector(2**eu_cmp_width-1 downto 0);
      eu_cmp_B    : in  std_ulogic_vector(2**eu_cmp_width-1 downto 0);
      eu_cmp_simd : in  std_ulogic_vector(eu_cmp_width-4 downto 0);
      eu_cmp_mode : in  std_ulogic_vector(2 downto 0);
      eu_cmp_Y    : out std_ulogic_vector(2**eu_cmp_width-1 downto 0);
      eu_cmp_Z    : out std_ulogic_vector(2**eu_cmp_width-1 downto 0)
      );
end eu_cmp;

-- Description of EU_CMP entity
architecture simple of eu_cmp is
   type cmp_array is array(eu_cmp_width-3 downto 0) of std_ulogic_vector(2**eu_cmp_width-1 downto 0);
   signal tree_in   : std_ulogic_vector(2**eu_cmp_width-1 downto 0);
   signal tree_out  : std_ulogic_vector(2**eu_cmp_width-1 downto 0);

   function inv_word(A : in std_ulogic_vector) return std_ulogic_vector is
      variable L : natural := A'low;
      variable H : natural := A'high;
      variable B : std_ulogic_vector(H downto L);

   begin
      for i in L to H loop
         B(i) := A(H+L-i);
      end loop;
      return B;
   end function;


   function cmp_chunk( A, B, CMP : in std_ulogic_vector ) return std_ulogic_vector is
      variable L : natural := A'length;
      variable K : natural := A'low;
      variable zero : std_ulogic;

   begin
      zero := CMP(K);

      case zero is
         when '1' => return A;
         when '0' => return B;
         when others => return A;
      end case;
   end function;

begin
   -- Component find_lsb
   tree_find_lsb : entity find_lsb
      generic map( find_lsb_width => eu_cmp_width )
      port map(
         find_lsb_input  => tree_in,
         find_lsb_output => tree_out,
         find_lsb_simd   => eu_cmp_simd
         );

   eu_cmp_in : process(eu_cmp_A, eu_cmp_B, eu_cmp_simd, eu_cmp_mode) is
      variable eu_cmp_not : std_ulogic_vector(2**eu_cmp_width-1 downto 0);

   begin
     -- Input of unit
     eu_cmp_not := not eu_cmp_A;
     case eu_cmp_mode is 
        when "000"  => tree_in <= inv_word(eu_cmp_A xor eu_cmp_B);
        when "001"  => tree_in <= inv_word(eu_cmp_A xor eu_cmp_B);
        when "010"  => tree_in <= inv_word(eu_cmp_A xor eu_cmp_B);
        when "011"  => tree_in <= inv_word(eu_cmp_A xor eu_cmp_B);
        when "100"  => tree_in <= inv_word(eu_cmp_A);
        when "101"  => tree_in <= inv_word(eu_cmp_not);
        when others => tree_in <= inv_word(eu_cmp_A);
     end case;
   end process;

   eu_inc_out : process(tree_out, eu_cmp_A, eu_cmp_B, eu_cmp_simd, eu_cmp_mode) is
      variable N : natural;
      variable L : natural := 2**eu_cmp_width;
      variable cmp_simd1 : cmp_array;
      variable cmp_simd2 : cmp_array;
      variable tree_cmp1 : std_ulogic_vector(2**eu_cmp_width-1 downto 0);
      variable tree_cmp2 : std_ulogic_vector(2**eu_cmp_width-1 downto 0);

   begin
      -- perform cmp tree
      for i in 0 to eu_cmp_width-3 loop
         N := 2**i;
         for j in 0 to N-1 loop
            cmp_simd1(i)((j+1)*L/N-1 downto j*L/N) := cmp_chunk(
               eu_cmp_A((j+1)*L/N-1 downto j*L/N),
               eu_cmp_B((j+1)*L/N-1 downto j*L/N),
               inv_word(tree_out((j+1)*L/N-1 downto j*L/N)));
            cmp_simd2(i)((j+1)*L/N-1 downto j*L/N) := cmp_chunk(
               eu_cmp_B((j+1)*L/N-1 downto j*L/N),
               eu_cmp_A((j+1)*L/N-1 downto j*L/N),
               inv_word(tree_out((j+1)*L/N-1 downto j*L/N)));
         end loop;
      end loop;

      -- search simd level and mux for cmp
      tree_cmp1 := cmp_simd1(0)(2**eu_cmp_width-1 downto 0);
      tree_cmp2 := cmp_simd2(0)(2**eu_cmp_width-1 downto 0);
      for i in 0 to eu_cmp_simd'high loop
         if eu_cmp_simd(i) = '1' then
            tree_cmp1 := cmp_simd1(i+1)(2**eu_cmp_width-1 downto 0);
            tree_cmp2 := cmp_simd2(i+1)(2**eu_cmp_width-1 downto 0);
         end if;
      end loop;

      -- Output of unit
      case eu_cmp_mode is
         when "000"  => eu_cmp_Y <= inv_word(tree_out); eu_cmp_Z <= inv_word(tree_out);
         when "001"  => eu_cmp_Y <= tree_cmp1; eu_cmp_Z <= tree_cmp1;
         when "010"  => eu_cmp_Y <= tree_cmp2; eu_cmp_Z <= tree_cmp2;
         when "011"  => eu_cmp_Y <= tree_cmp1; eu_cmp_Z <= tree_cmp2;
         when "100"  => eu_cmp_Y <= inv_word(tree_out); eu_cmp_Z <= inv_word(tree_out);
         when "101"  => eu_cmp_Y <= inv_word(tree_out); eu_cmp_Z <= inv_word(tree_out);
         when others => eu_cmp_Y <= inv_word(tree_out); eu_cmp_Z <= inv_word(tree_out);
      end case;
   end process;
end;
