--------------------------------------------------------------------------
-- File vhdl/eu_inc/find_lsb.vhdl - Binary tree for find LSB
-- Copyright (C) 2002 Etienne Labarre (etienne.labarre@gadz.org)
--
-- Sun Jun 30 11:37:34 CEST 2002 : E. LABARRE
--  First version
--
-- Mon Jul  8 19:56:53 CEST 2002 : E. LABARRE
--  Change to sequential description
--
-- Sat Jul 27 18:04:48 CEST 2002 : E. LABARRE
--  Use of reduce_and function (library std_1164)
--
--------------------------BEGIN-VHDL-LICENCE-----------------------------
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
---------------------------END-VHDL-LICENCE------------------------------
--
-- Binary tree for find lsb in a word. Size 2**N, with N = max 8.
--
-- fbit_simd : Perform SIMD operation. 
--
-- fbit_simd = "...00000" => N    bits operations
-- fbit_simd = "...00001" => N/2  bits operations
-- fbit_simd = "...00011" => N/4  bits operations
-- etc... with chunk size mini = 8 bits
-- 
--------------------------------------------------------------------------

-- include standard libraries
LIBRARY ieee;
USE ieee.std_logic_1164.ALL;
USE ieee.std_logic_misc.ALL;

-- Definition of find_lsb entity
entity find_lsb is
   generic( find_lsb_width : natural := 6 );
   port (
      find_lsb_input  : in  std_ulogic_vector(2**find_lsb_width-1 downto 0);
      find_lsb_output : out std_ulogic_vector(2**find_lsb_width-1 downto 0);
      find_lsb_simd   : in  std_ulogic_vector(find_lsb_width-4 downto 0)
      );
end find_lsb;


-- Description of find_lsb entity
architecture simple of find_lsb is

   constant BITNB   : natural := 2**find_lsb_width;
   type simd_array  is array(find_lsb_width-3 downto 0) of std_ulogic_vector(BITNB-1 downto 0);
   type inter_array  is array(BITNB-1 downto 0) of std_ulogic_vector(64 downto 0);

   
   function shiftl1 (A : in std_ulogic_vector) return std_ulogic_vector is
      constant L : natural := A'length;
      variable aa, bb : std_ulogic_vector(L-1 downto 0);

   begin
      --pragma synthesis_off
      assert L > 1;
      --pragma synthesis_on
      aa := A;
      bb(0) := '1';
      bb(L-1 downto 1) := aa(L-2 downto 0);
      return bb;
   end function;

begin
   and_tree : process(find_lsb_input,find_lsb_simd) is
      variable aa    : std_ulogic_vector(BITNB-1 downto 0);
      variable bb    : inter_array;
      variable cc    : simd_array;
      variable dd    : simd_array;
      variable k     : natural;

   begin
      -- First stage : cascade_and for 8 bits chunks
      for i in 0 to 2**(find_lsb_width-3)-1 loop
         for j in 0 to 7 loop
	    aa(8*i+j) := and_reduce(find_lsb_input(8*i+j downto 8*i));
         end loop;
      end loop;

      -- Second stage : rename variables
      for i in 0 to BITNB-1 loop
         bb(i)(0) := aa(i);
	 if (i>7) then
	    for l in 1 to (i/8) loop
	       k := 8*(i/8) - 8*(l-1);
	       bb(i)(l) := aa(k-1);
	    end loop;
	 end if;
      end loop;
    
      -- Third stage : cascade_and for all SIMD outputs (8-16-32-64 bits)
      for i in 0 to find_lsb_width-3 loop
         for j in 0 to (2**i)-1 loop
            for k in 0 to BITNB/(2**i)-1 loop
	       cc(i)(j*BITNB/(2**i)+k) := and_reduce(bb(j*BITNB/(2**i)+k)(k/8 downto 0));
	    end loop;
	 end loop;
      end loop;

      -- Fourth stage : shift to left 1 bit for all chunks
      for i in 0 to find_lsb_width-3 loop
         for j in 0 to (2**i)-1 loop
	    dd(i)((j+1)*BITNB/(2**i)-1 downto j*BITNB/(2**i)) := 
	       shiftl1(cc(i)((j+1)*BITNB/(2**i)-1 downto j*BITNB/(2**i)));
	 end loop;
      end loop;

      -- Fifth stage : choice of output
      find_lsb_output <= dd(0)(BITNB-1 downto 0);
      for i in 0 to find_lsb_simd'high loop
         if find_lsb_simd(i) = '1' then
	    find_lsb_output <= dd(i+1)(BITNB-1 downto 0);
	 end if;
      end loop;

   end process;
end;
